/* Monitor */
#include <pthread.h>
#include <stdint.h>

/* Type of compose function that accumulate partial results */
typedef uintmax_t (*compose_t)(uintmax_t y,uintmax_t r) ;


typedef struct {
  int nrun ;       /* number of tasks being executed */
  int ncompleted ; /* number if tasks being completed */
  int waiting ;    /* flag set if master is waiting  */
  pthread_mutex_t *lock;
  pthread_cond_t *cond ;
  compose_t compose ; /* compose function */
  uintmax_t r ;    /* result of computation */

} monitor_t ;

/* To be called by master with result initial value */
monitor_t *alloc_monitor(compose_t f,uintmax_t init_t) ;

void free_monitor(monitor_t *p) ;

/*
   To be called by worker :
   1. Pass partial result y, so as to update result of computation
      [m->r = m->compose(y,m->r)]
   2. Signal a taks is completed
*/
void leave_monitor(monitor_t *m,uintmax_t y) ;

/* To be called by master to signal a task is being executed */
void enter_monitor(monitor_t *m) ;

/* To be called by master to wait for ntasks being completed.
   returns computation result */
uintmax_t wait_monitor(monitor_t *m,int ntasks) ;
