#include <stdlib.h>
#include "pool.h"

/* Pool */

pool_t *alloc_pool(int maxrun) {
  pool_t *p = malloc_check(sizeof(*p)) ;
  p->maxrun = maxrun ;
  p->nrun = 0 ;
  p->waiting = 0 ;
  p->lock = alloc_mutex() ;
  p->cond = alloc_cond() ;  
  return p ;
}

void free_pool(pool_t *p) {
  free_mutex(p->lock) ;
  free_cond(p->cond) ;
  free(p) ;
}

void leave_pool(pool_t *p) {
  lock_mutex(p->lock) ;
  p->nrun-- ;
  if (p->waiting) signal_cond(p->cond) ;
  unlock_mutex(p->lock) ;
}

void look_pool(pool_t *p) {
  lock_mutex(p->lock) ;
  p->waiting = 1 ;
  while (p->nrun >= p->maxrun) { 
    wait_cond(p->cond,p->lock) ;
  }
  p->waiting = 0 ;
  p->nrun++ ;
  unlock_mutex(p->lock) ;
}
