#include <stdlib.h>
#include <stdio.h>
#include "basic.h"


static int verbose = 0 ;

/* Semaphore */

typedef struct {
  volatile int nfree ;
  pthread_mutex_t *mutex ;
  pthread_cond_t *cond ;
} semaphore_t ;


semaphore_t *alloc_semaphore(int capacity) {
  semaphore_t *p = malloc_check(sizeof(*p)) ;
  p->nfree = capacity ;
  p->mutex = alloc_mutex() ;
  p->cond = alloc_cond() ;
  return p ;
}

void free_semaphore(semaphore_t *p) {
  free_mutex(p->mutex) ;
  free_cond(p->cond) ;
  free(p) ;
}

void wait_semaphore(semaphore_t *p) {
  lock_mutex(p->mutex) ;
  while (p->nfree <= 0) wait_cond(p->cond,p->mutex) ;
  p->nfree-- ;
  unlock_mutex(p->mutex) ;
}

void post_semaphore(semaphore_t *p) {
  lock_mutex(p->mutex) ;
  p->nfree++ ;
  broadcast_cond(p->cond) ;
  unlock_mutex(p->mutex) ;
}

/* Threads */

typedef struct {
  int size ;
  pthread_barrier_t *b ;
  semaphore_t *sem ;
} common_t ;

typedef struct {
  int id ;
  common_t *common ;
} ctx_t ;


void *T1(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  for (int k = q->size-1 ; k >= 0 ; k--) {
    wait_semaphore(q->sem) ;
    printf("+") ;
    delay(50) ;
    printf("-") ;
    post_semaphore(q->sem) ;
    wait_barrier(q->b) ;
    if (p->id == 0) printf("\n") ;
    wait_barrier(q->b) ;
  }
  return NULL ;
}

void run(int size, int nprocs, int m) {
  common_t c ;
  c.b = alloc_barrier(nprocs) ;
  c.sem = alloc_semaphore(m) ;
  c.size = size ;

  pthread_t th[nprocs] ;
  ctx_t a[nprocs] ;
  for (int k = 0 ; k < nprocs ; k++) {
    ctx_t *p = &a[k] ;
    p->id = k ; p->common = &c ;
    create_thread(&th[k],T1,p) ;
  }
  for (int k = 0 ; k < nprocs ; k++) (void)join_thread(&th[k]) ;
  free_semaphore(c.sem) ;
  free_barrier(c.b) ;
}


void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [size] [nproc] [capacity]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int size = 10 ;
  int n = 4 ;
  int m = 2 ;
  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    size = atoi(argv[0]) ;
  } else if (argc == 2) {
    size = atoi(argv[0]) ;
    n = atoi(argv[1]) ;
  } else if (argc > 2) {
    size = atoi(argv[0]) ;
    n = atoi(argv[1]) ;
    m = atoi(argv[2]) ;
  }
  run(size,n,m) ;
  return 0 ;
}
